# Discover include directory of Cyantities as a script.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Import Cyantities, get path to its root, and add the path of the
# _cpp include directory:
from pathlib import Path
import cyantities

def cyantities_print_include():
    root = Path(cyantities.__file__).resolve().parent

    include = root / "_cpp"

    print(str(include))


# Quantities with units.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.


import numpy as np
from cython.cimports.cpython.ref cimport PyObject, PyTypeObject
from cpython.object cimport PyObject_TypeCheck
from numpy cimport ndarray, float64_t, PyArrayObject, npy_intp,\
    NPY_DOUBLE
from .errors import UnitError
from .unit cimport CppUnit, Unit, parse_unit, generate_from_cpp, format_unit
from .quantity cimport Quantity
from libc.math cimport log10
from libc.stdint cimport int16_t
from libcpp cimport bool

cdef extern from *:
    """
    #include <cstdlib>
    #include <numpy/ndarraytypes.h>

    bool is_single_stride(
        npy_intp* stride,
        npy_intp* shape,
        int ndim,
        npy_intp itemsize
    )
    {
        npy_intp expected_stride = itemsize;
        for (int i=0; i<ndim; ++i){
            int j = ndim - i - 1;
            auto x = std::div(stride[j], expected_stride);
            if ((x.quot != 1) || (x.rem != 0)){
                return false;
            }

            /* New expected stride: */
            expected_stride *= shape[j];
        }

        return true;
    }

    size_t ptr2int(const char* ptr){
        return (size_t)ptr;
    }
    """
    npy_intp* PyArray_SHAPE(PyArrayObject*)
    npy_intp* PyArray_STRIDES(PyArrayObject*)
    void* PyArray_DATA(PyArrayObject*)
    int PyArray_NDIM(PyArrayObject*)
    npy_intp PyArray_ITEMSIZE(PyArrayObject*)
    npy_intp PyArray_SIZE(PyArrayObject*)
    PyTypeObject PyArray_Type
    cppclass PyArray_Descr
    int PyArray_CastScalarToCtype(PyObject*, void*, PyArray_Descr*)
    PyArray_Descr* PyArray_DescrFromType(int typenum)


    size_t ptr2int(const char* ptr)
    bool is_single_stride(
        npy_intp* stride,
        npy_intp* shape,
        int ndim,
        npy_intp itemsize
    )


#
# Dummy buffer:
#
cdef double[1] dummy_double
dummy_double[0] = 1.938928939273982423e-78


#
# Double:
#
cdef PyArray_Descr* _DOUBLE_ARRAY_TYPE = PyArray_DescrFromType(NPY_DOUBLE)


cdef Quantity _multiply_quantities(Quantity q0, Quantity q1):
    """
    Multiply two quantities.
    """
    cdef Quantity res = Quantity.__new__(Quantity)
    cdef CppUnit unit = q0._unit * q1._unit

    if q0._is_scalar and q1._is_scalar:
        res._cyinit(True, q0._val * q1._val, None, unit)

    elif q0._is_scalar:
        if q0._val == 1.0:
            # Shortcut: Do not copy.
            res._cyinit(
                False, dummy_double[0], q1._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0], float(q0._val) * q1._val_object, unit
            )

    elif q1._is_scalar:
        if q1._val == 1.0:
            # Shortcut: Do not copy.
            res._cyinit(
                False, dummy_double[0], q0._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0], float(q1._val) * q0._val_object, unit
            )

    else:
        res._cyinit(
            False, dummy_double[0], q0._val_object * q1._val_object, unit
        )

    return res


cdef Quantity _divide_quantities(Quantity q0, Quantity q1):
    """
    Multiply two quantities.
    """
    cdef Quantity res = Quantity.__new__(Quantity)
    cdef CppUnit unit = q0._unit / q1._unit

    if q0._is_scalar and q1._is_scalar:
        res._cyinit(True, q0._val / q1._val, None, unit)

    elif q0._is_scalar:
        res._cyinit(
            False, dummy_double[0], float(q0._val) / q1._val_object, unit
        )

    elif q1._is_scalar:
        if q1._val == 1.0:
            # Shortcut: Do not copy.
            res._cyinit(
                False, dummy_double[0], q0._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0], q0._val_object / float(q1._val), unit
            )

    else:
        res._cyinit(
            False, dummy_double[0], q0._val_object / q1._val_object, unit
        )

    return res


cdef Quantity _add_quantities_equal_scale(Quantity q0, double s0, Quantity q1,
                                          double s1, CppUnit unit):
    """
    Adds two quantities of equal scale.
    """
    cdef Quantity res = Quantity.__new__(Quantity)
    s0 *= (q0._unit / unit).total_scale()
    s1 *= (q1._unit / unit).total_scale()

    if q0._is_scalar and q1._is_scalar:
        res._cyinit(True, s0 * q0._val + s1 * q1._val, None, unit)

    elif q0._is_scalar:
        if s1 == 1.0:
            res._cyinit(
                False, dummy_double[0],
                float(s0 * q0._val) + q1._val_object, unit
            )
        elif s1 == -1.0:
            res._cyinit(
                False, dummy_double[0],
                float(s0 * q0._val) - q1._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0],
                float(s0 * q0._val) + s1 * q1._val_object, unit
            )

    elif q1._is_scalar:
        if s0 == 1.0:
            res._cyinit(
                False, dummy_double[0],
                float(s1 * q1._val) + q0._val_object, unit
            )
        elif s0 == -1.0:
            res._cyinit(
                False, dummy_double[0],
                float(s1 * q1._val) - q0._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0],
                float(s1 * q1._val) + s0 * q0._val_object, unit
            )

    else:
        if s0 == 1.0 and s1 == 1.0:
            res._cyinit(
                False, dummy_double[0], q0._val_object + q1._val_object, unit
            )
        elif s0 == 1.0 and s1 == -1.0:
            res._cyinit(
                False, dummy_double[0], q0._val_object - q1._val_object, unit
            )
        elif s0 == 1.0:
            res._cyinit(
                False, dummy_double[0],
                q0._val_object + s1 * q1._val_object, unit
            )
        elif s0 == -1.0:
            res._cyinit(
                False, dummy_double[0],
                s1 * q1._val_object - q0._val_object, unit
            )
        elif s1 == 1.0:
            res._cyinit(
                False, dummy_double[0],
                s0 * q0._val_object + q1._val_object, unit
            )
        elif s1 == -1.0:
            res._cyinit(
                False, dummy_double[0],
                s0 * q0._val_object - q1._val_object, unit
            )
        else:
            res._cyinit(
                False, dummy_double[0],
                s0 * q0._val_object + s1 * q1._val_object, unit
            )

    return res


cdef Quantity _add_quantities(Quantity q0, Quantity q1):
    """
    Add two quantities.
    """
    if q0._unit == q1._unit:
        return _add_quantities_equal_scale(q0, 1.0, q1, 1.0, q0._unit)

    # Otherwise need to decide which unit to add in:
    cdef CppUnit scale = q0._unit / q1._unit
    if not scale.dimensionless():
        raise RuntimeError("Cannot add quantities of different dimension.")

    # Decide which unit to add in:
    cdef int16_t dec_exp = scale.decadal_exponent()
    cdef double total_exp = dec_exp + log10(scale.conversion_factor())
    if total_exp <= 0:
        # Use scale of q1.
        return _add_quantities_equal_scale(q0, 1.0, q1, 1.0, q1._unit)
    else:
        # Use scale of q0.
        return _add_quantities_equal_scale(q0, 1.0, q1, 1.0, q0._unit)


cdef Quantity _subtract_quantities(Quantity q0, Quantity q1):
    """
    Add two quantities.
    """
    if q0._unit == q1._unit:
        return _add_quantities_equal_scale(q0, 1.0, q1, -1.0, q0._unit)

    # Otherwise need to decide which unit to add in:
    cdef CppUnit scale = q0._unit / q1._unit
    if not scale.dimensionless():
        raise RuntimeError("Cannot add quantities of different dimension.")

    # Decide which unit to add in:
    cdef int16_t dec_exp = scale.decadal_exponent()
    cdef double total_exp = dec_exp + log10(scale.conversion_factor())
    if total_exp <= 0:
        # Use scale of q1.
        return _add_quantities_equal_scale(q0, 1.0, q1, -1.0, q1._unit)
    else:
        # Use scale of q0.
        return _add_quantities_equal_scale(q0, 1.0, q1, -1.0, q0._unit)


cdef Quantity _power_quantity(Quantity q0, int b):
    """
    Compute the power of a quantity.
    """
    cdef CppUnit unit = q0._unit.power(b)
    cdef Quantity res = Quantity.__new__(Quantity)
    if q0._is_scalar:
        res._cyinit(True, q0._val ** b, None, unit)

    else:
        res._cyinit(False, dummy_double[0], q0._val_object ** b, unit)

    return res


cdef class Quantity:
    """
    A physical quantity: a single or array of real numbers with an associated
    physical unit.
    """
    __dict__: dict

    def __init__(self, value, unit, bool copy=True):
        #
        # First determine the values (scalar / ndarray)
        # and setup all corresponding members for a call
        # to _cyinit
        #
        cdef double d_value
        cdef bool is_scalar
        cdef object val_object
        if isinstance(value, float) or isinstance(value, int):
            is_scalar = True
            d_value = value
            val_object = None
        elif isinstance(value, np.ndarray):
            is_scalar = False
            d_value = dummy_double[0]
            if copy:
                val_object = value.copy()
                val_object.flags['WRITEABLE'] = False
            else:
                val_object = value
        else:
            raise TypeError("'value' has to be either a float or a NumPy array.")

        #
        # Then the unit:
        #
        cdef Unit unit_Unit
        cdef CppUnit cpp_unit
        if isinstance(unit, Unit):
            unit_Unit = unit
            cpp_unit = unit_Unit._unit
        elif isinstance(unit, str):
            cpp_unit = parse_unit(unit)

        else:
            raise TypeError("'unit' has to be either a string or a Unit.")

        self._cyinit(is_scalar, d_value, val_object, cpp_unit)


    cdef _cyinit(self, bool is_scalar, double val, object val_object,
                 CppUnit unit):
        if self._initialized:
            raise RuntimeError("Trying to initialize a second time.")
        self._is_scalar = is_scalar
        self._val = val
        cdef double[::1] buffer
        cdef PyArrayObject* pao
        cdef npy_intp* shape
        cdef npy_intp* strides
        cdef npy_intp itemsize
        cdef int ndim
        cdef PyObject* val_array_ptr
        if isinstance(val_object, np.ndarray):
            # Ensure that the array is of double type and ensure that the
            # underlying buffer is non-strided.
            # This single Python call also ensures that non-ndarray types
            # can be handled.
            val_array = np.array(
                val_object, dtype=np.double, copy=False, order='C'
            )
            val_array_ptr = <PyObject*>val_array

            # Obtain a PyArrayObject pointer and get all the relevant info:
            pao = <PyArrayObject*>val_array_ptr
            ndim = PyArray_NDIM(pao)
            shape = PyArray_SHAPE(pao)
            strides = PyArray_STRIDES(pao)
            itemsize = PyArray_ITEMSIZE(pao)

            # Sanity: Assert that all is single stride:
            if not is_single_stride(strides, shape, ndim, itemsize):
                raise RuntimeError(
                    "Could not get a non-strided version of the input "
                    "array."
                )

            # Now set the attributes.
            # We keep an explicit reference to the NDArray 'val_array' so as
            # to have automatic reference counting.
            self._val_object = val_array
            self._val_array_ptr = <double*>PyArray_DATA(pao)
            self._val_array_N = PyArray_SIZE(pao)
        else:
            self._val_object = None
            self._val_array_ptr = NULL
            self._val_array_N = 0
        self._unit = unit

        # Add, if dimensionless, the __array__ routine:
        if unit.dimensionless():
            self.__array__ = self._array

        self._initialized = True


    def __float__(self):
        """
        Returns, if dimensionally possible, a scalar.
        """
        if not self._unit.dimensionless():
            raise RuntimeError("Attempting to convert a dimensional quantity "
                               "to dimensionless scalar.")
        if not self._is_scalar:
            raise RuntimeError("Attempting to convert a non-scalar quantity to "
                               "a scalar.")

        return float(self._val * self._unit.total_scale())


    def _array(self) -> np.ndarray:
        """
        Returns, if dimensionally possible, a scalar array.
        """
        # TODO: More arguments of the __array__ routine protocol of numpy!
        if not self._unit.dimensionless():
            raise RuntimeError("Attempting to get array of a dimensional quantity.")
        cdef object array
        cdef double scale = self._unit.total_scale()
        if self._is_scalar:
            return np.full(1, self._val * scale)

        if scale != 1.0:
            return self._val_object * float(scale)

        return self._val_object.view()


    def __repr__(self) -> str:
        """
        String representation.
        """
        cdef str rep = "Quantity("
        if self._is_scalar:
            rep += str(float(self._val))
        else:
            rep += self._val_object.__repr__()
        rep += ", '"
        rep += format_unit(self._unit, 'coherent')
        rep += "')"

        return rep


    def __mul__(self, other):
        """
        Multiply this quantity with another quantity or float.
        """
        # Classifying the other object:
        cdef Quantity other_quantity
        cdef Unit a_unit

        #
        # Initialize the quantity that we would like to multiply with:
        #
        if isinstance(other, np.ndarray):
            other_quantity = Quantity.__new__(Quantity)
            if other.size == 1:
                other_quantity._cyinit(True, other.flat[0], None, CppUnit())
            else:
                other_quantity._cyinit(False, dummy_double[0], other, CppUnit())

        elif isinstance(other, float) or isinstance(other, int):
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, other, None, CppUnit())

        elif isinstance(other, Quantity):
            other_quantity = other

        elif isinstance(other, Unit):
            a_unit = other
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, 1.0, None, a_unit._unit)
        else:
            return NotImplemented

        return _multiply_quantities(self, other_quantity)


    def __rmul__(self, other):
        """
        Multiply this quantity with another quantity or float (from the
        right).
        """
        # Classifying the other object:
        cdef Quantity other_quantity
        cdef Unit a_unit

        #
        # Initialize the quantity that we would like to multiply with:
        #
        if isinstance(other, np.ndarray):
            other_quantity = Quantity.__new__(Quantity)
            if other.size == 1:
                other_quantity._cyinit(True, other.flat[0], None, CppUnit())
            else:
                other_quantity._cyinit(False, dummy_double[0], other, CppUnit())

        elif isinstance(other, float) or isinstance(other, int):
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, other, None, CppUnit())

        elif isinstance(other, Quantity):
            other_quantity = other

        elif isinstance(other, Unit):
            a_unit = other
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, 1.0, None, a_unit._unit)
        else:
            return NotImplemented

        return _multiply_quantities(other_quantity, self)


    def __truediv__(self, other):
        """
        Divide this quantity by another quantity or float.
        """
        # Classifying the other object:
        cdef Quantity other_quantity
        cdef Unit a_unit

        #
        # Initialize the quantity that we would like to multiply with:
        #
        if isinstance(other, np.ndarray):
            other_quantity = Quantity.__new__(Quantity)
            if other.size == 1:
                other_quantity._cyinit(True, other.flat[0], None, CppUnit())
            else:
                other_quantity._cyinit(False, dummy_double[0], other, CppUnit())

        elif isinstance(other, float) or isinstance(other, int):
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, other, None, CppUnit())

        elif isinstance(other, Quantity):
            other_quantity = other

        elif isinstance(other, Unit):
            a_unit = other
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, 1.0, None, a_unit._unit)
        else:
            return NotImplemented

        return _divide_quantities(self, other_quantity)


    def __rtruediv__(self, other):
        """
        Divide this quantity by another quantity or float.
        """
        # Classifying the other object:
        cdef Quantity other_quantity
        cdef Unit a_unit

        #
        # Initialize the quantity that we would like to multiply with:
        #
        if isinstance(other, np.ndarray):
            other_quantity = Quantity.__new__(Quantity)
            if other.size == 1:
                other_quantity._cyinit(True, other.flat[0], None, CppUnit())
            else:
                other_quantity._cyinit(False, dummy_double[0], other, CppUnit())

        elif isinstance(other, float) or isinstance(other, int):
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, other, None, CppUnit())

        elif isinstance(other, Quantity):
            other_quantity = other

        elif isinstance(other, Unit):
            a_unit = other
            other_quantity = Quantity.__new__(Quantity)
            other_quantity._cyinit(True, 1.0, None, a_unit._unit)
        else:
            return NotImplemented

        return _divide_quantities(other_quantity, self)


    def __add__(self, Quantity other):
        if not self._unit.same_dimension(other._unit):
            raise UnitError("Trying to add two quantities of incompatible "
                            "units.")

        return _add_quantities(self, other)


    def __sub__(self, Quantity other):
        if not self._unit.same_dimension(other._unit):
            raise UnitError("Trying to subtract two quantities of incompatible "
                            "units.")

        return _subtract_quantities(self, other)


    def __pow__(self, exponent):
        if not isinstance(exponent, int):
            raise TypeError("Quantities can be exponentiated only to integer "
                "powers."
            )
        return _power_quantity(self, exponent)


    def __neg__(self):
        """
        Unary negative.
        """
        # Mostly a copy, we just have to see which part of the value
        # (scalar or ndarray?) we have to negate:
        cdef Quantity res = Quantity.__new__(Quantity)
        if self._is_scalar:
            res._cyinit(
                True, -self._val, None, self._unit
            )
        else:
            res._cyinit(
                False, dummy_double[0], -self._val_object, self._unit
            )

        return res


    def __abs__(self):
        """
        Unary absolute value.
        """
        # Mostly a copy, we just have to see which part of the value
        # (scalar or ndarray?) we have to negate:
        cdef Quantity res = Quantity.__new__(Quantity)
        if self._is_scalar:
            res._cyinit(
                True, abs(self._val), None, self._unit
            )
        else:
            res._cyinit(
                False, dummy_double[0], np.abs(self._val_object), self._unit
            )

        return res


    def __eq__(self, other):
        # First the case that the other is not a Quantity.
        # This results in nonzero only if this quantity is
        # dimensionless:
        if not isinstance(other, Quantity):
            if not self._unit.dimensionless():
                return False
            if self._is_scalar:
                return float(self._val) == other
            return self._val_object == other

        # Now compare quantities:
        cdef Quantity oq = other
        if not self._unit.same_dimension(oq._unit):
            return False

        # Check whether there's a scale difference:
        cdef CppUnit div_unit = self._unit / oq._unit
        cdef double scale = div_unit.total_scale()
        if scale == 1.0:
            # No scale difference. Make the two possible
            # comparisons:
            if self._is_scalar and oq._is_scalar:
                return self._val == oq._val
            elif not self._is_scalar and not oq._is_scalar:
                return self._val_object == oq._val_object
            return False

        # Have scale difference. Make the two possible
        # comparisons:
        if self._is_scalar and oq._is_scalar:
            return self._val == scale*oq._val
        elif not self._is_scalar and not oq._is_scalar:
            return self._val_object == scale * oq._val_object
        return False


    def __getitem__(self, index) -> Quantity:
        """
        Indexing of this quantity.
        """
        if self._is_scalar:
            raise IndexError("Cannot index a scalar Quantity.")

        # Here we use NumPy indexing to process all the indexing
        # pecularities and throw appropriate errors (which we then
        # only augment by some decoration indicating that it's a
        # Quantity error):
        cdef object val_object
        try:
            val_object = self._val_object[index]
        except IndexError as ie:
            raise IndexError("Quantity index error: " + ie.args[0])


        # Initialize the new Quantity:
        cdef Quantity q = Quantity.__new__(Quantity)
        cdef PyObject* val_object_ptr
        cdef double* data
        val_object_ptr = <PyObject*>val_object
        cdef bool is_array = PyObject_TypeCheck(
            val_object, &PyArray_Type
        )
        cdef double val
        cdef int success
        if not is_array:
            val = 0.0
            success = PyArray_CastScalarToCtype(
                val_object_ptr, &val, _DOUBLE_ARRAY_TYPE
            )
            if success != 0:
                raise RuntimeError(
                    "Could not cast index result scalar to double. This error "
                    "should not occur."
                )
            q._cyinit(True, val, None, self._unit)

        else:
            q._cyinit(False, dummy_double[0], val_object, self._unit)

        return q


    def shape(self) -> int | tuple[int,...]:
        """
        Return this quantity's shape.
        """
        if self._is_scalar:
            return 1
        return self._val_object.shape


    def unit(self) -> Unit:
        return generate_from_cpp(self._unit)


    cdef QuantityWrapper wrapper(self) nogil:
        """
        Return a QuantityWrapper instance for talking to C++.
        """
        if self._is_scalar:
            return QuantityWrapper(self._val, self._unit)
        else:
            return QuantityWrapper(
                self._val_array_ptr,
                self._val_array_N,
                self._unit)


    @staticmethod
    cdef Quantity zeros_like(Quantity other, object unit):
        """
        Returns a zero-value quantity with shape like another,
        potentially with a different unit.
        """
        # First check the unit:
        cdef Unit src_unit
        cdef CppUnit dest_unit
        if unit is None:
            dest_unit = other._unit
        elif isinstance(unit, str):
            dest_unit = parse_unit(unit)
        elif isinstance(unit, Unit):
            src_unit = unit
            dest_unit = src_unit._unit
        else:
            raise TypeError("'unit' must be a Unit instance, unit-specifying "
                "string, or None."
            )

        # Now determine the shape of the target quantity:
        cdef Quantity res = Quantity.__new__(Quantity)
        if other._is_scalar:
            res._cyinit(True, 0.0, None, dest_unit)

        else:
            # Generate a NumPy array with shape equal to the other
            # quantity:
            res._cyinit(False, dummy_double[0],
                np.zeros_like(other._val_object),
                dest_unit
            )

        return res


# Solves a parabola with friction.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import numpy as np
cimport numpy as cnp
from cyantities.unit cimport parse_unit, CppUnit
from cyantities.quantity cimport Quantity, QuantityWrapper


cdef extern from "parasolve.hpp":
    void solve_ball_throw_with_friction(
        const QuantityWrapper& t0_qw,
        const QuantityWrapper& dt0_qw,
        const QuantityWrapper& x0_qw,
        const QuantityWrapper& y0_qw,
        const QuantityWrapper& vx0_qw,
        const QuantityWrapper& vy0_qw,
        const QuantityWrapper& cw_qw,
        const QuantityWrapper& r_qw,
        const QuantityWrapper& rho_qw,
        const QuantityWrapper& rho_air_qw,
        const QuantityWrapper& t_qw,
        QuantityWrapper& x_qw,
        QuantityWrapper& y_qw,
        double err_rel,
        double err_abs
    ) except+




def ball_throw_with_friction(
        Quantity t, Quantity x0, Quantity y0, Quantity vx0, Quantity vy0,
        Quantity r, Quantity cw, Quantity rho, Quantity rho_air,
        double err_rel = 1e-6, double err_abs = 1e-6
    ):
    """

    """
    # Make sure that all parameters are scalar:
    assert (x0._is_scalar and y0._is_scalar and vx0._is_scalar
            and vy0._is_scalar and r._is_scalar and cw._is_scalar
            and rho._is_scalar and rho_air._is_scalar)

    # Number of time steps:
    cdef size_t Nt
    if t._is_scalar:
        Nt = 1
    else:
        Nt = t._val_ndarray.size

    cdef CppUnit meter = parse_unit("m")
    cdef CppUnit seconds = parse_unit("s")
    cdef Quantity x = Quantity.__new__(Quantity)
    cdef cnp.ndarray xarr = np.empty(Nt)
    x._cyinit(False, np.NaN, xarr, meter)
    cdef Quantity y = Quantity.__new__(Quantity)
    cdef cnp.ndarray yarr = np.empty(Nt)
    y._cyinit(False, np.NaN, yarr, meter)

    cdef Quantity t0 = Quantity.__new__(Quantity)
    t0._cyinit(True, 0.0, None, seconds)
    cdef Quantity dt0 = Quantity.__new__(Quantity)
    dt0._cyinit(True, 1e-3, None, seconds)

    solve_ball_throw_with_friction(
        t0.wrapper(), dt0.wrapper(), x0.wrapper(), y0.wrapper(),
        vx0.wrapper(), vy0.wrapper(), cw.wrapper(), r.wrapper(),
        rho.wrapper(), rho_air.wrapper(), t.wrapper(),
        x.wrapper(), y.wrapper(), err_rel, err_abs
    )

    return x, y


# SI Units in Python, accelerated by Cython.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from libcpp.vector cimport vector
from libc.stdint cimport uint8_t, int8_t, int16_t
from libcpp cimport bool

from .unit cimport base_unit_t, UnitBuilder, CppUnit, Unit, base_unit_array_t,\
                   base_unit_index_t


#
# Some convenience functions:
#
cdef extern from * namespace "cyantities" nogil:
    """
    namespace cyantities {
    static base_unit_t _base_unit_from_index(base_unit_index_t i)
    {
        return static_cast<base_unit_t>(i);
    }
    }
    """
    base_unit_t _base_unit_from_index(base_unit_index_t i) nogil



#######################################################################
#
#                         Cython unit parsing.
#
#######################################################################

cdef void _parse_unit_single(str unit, int prefix, int exponent,
                             UnitBuilder& builder):
    """
    A single unit's representation to SI basis units.
    """
    #
    # Prefixes:
    #
    cdef bool has_prefix = False
    cdef size_t n = len(unit)
    cdef Py_UCS4 first
    if n == 0:
        raise RuntimeError("Empty unit string.")
    elif n > 1:
        # Parse prefies:
        first = unit[0]
        if first == u'T':
            # Tera.
            builder.add_decadal_exponent(12 * prefix * exponent)
            has_prefix = True
        elif first == u'G':
            # Giga-prefix overlaps with Gray.
            # Catch that overlap here:
            if n == 2 and unit == 'Gy':
                builder.add_base_unit_occurrence(SI_METER,   2 * prefix * exponent)
                builder.add_base_unit_occurrence(SI_SECOND, -2 * prefix * exponent)
                return
            # Giga.
            builder.add_decadal_exponent(9 * prefix * exponent)
            has_prefix = True
        elif first == u'M':
            # Mega
            builder.add_decadal_exponent(6 * prefix * exponent)
            has_prefix = True
        elif first == u'k':
            # kilo-prefix overlaps with kilogram.
            # Catch that overlap here:
            if n == 2 and unit == "kg":
                builder.add_base_unit_occurrence(SI_KILOGRAM, prefix * exponent)
                return
            # kilo-prefix overlaps with katal:
            elif n == 3 and unit == "kat":
                builder.add_base_unit_occurrence(SI_MOLE,    1 * prefix * exponent)
                builder.add_base_unit_occurrence(SI_SECOND, -1 * prefix * exponent)
                return

            # kilo
            builder.add_decadal_exponent(3 * prefix * exponent)
            has_prefix = True
        elif first == u'h':
            # hecto
            builder.add_decadal_exponent(2 * prefix * exponent)
            has_prefix = True
        elif first == u'd':
            # dezi
            builder.add_decadal_exponent(-1 * prefix * exponent)
            has_prefix = True
        elif first == u'c':
            # centi-prefix overlaps with Candela.
            # Catch that overlap here:
            if n == 2 and unit == "cd":
                builder.add_base_unit_occurrence(SI_CANDELA, prefix * exponent)
                return
            # centi
            builder.add_decadal_exponent(-2 * prefix * exponent)
            has_prefix = True
        elif first == u'm':
            # milli-prefix overlaps with mol.
            # Catch that overlap here:
            if n == 3 and unit == "mol":
                builder.add_base_unit_occurrence(SI_MOLE, prefix * exponent)
                return
            # milli
            builder.add_decadal_exponent(-3 * prefix * exponent)
            has_prefix = True
        elif first == u'µ':
            # micro
            builder.add_decadal_exponent(-6 * prefix * exponent)
            has_prefix = True
        elif first == u'n':
            # nano
            builder.add_decadal_exponent(-9 * prefix * exponent)
            has_prefix = True
        elif first == u'p':
            # pico
            builder.add_decadal_exponent(-12 * prefix * exponent)
            has_prefix = True

        # Now need to cut the prefix:
        if has_prefix:
            unit = unit[1:]

    # A note for future performance optimization:
    # This function does a lot of string comparisons. Potentially, this process
    # might be sped up using a dict lookup for 'unit' (but that's also not clear
    # - measurement would be key).
    # Probably, units will not be parsed often enough in typical code to incur a
    # significant performance overhead, so that this function should be OK,
    # and simplicity in writing this function takes precendence.

    #
    # SI base units:
    #
    if unit == "m":
        builder.add_base_unit_occurrence(SI_METER, prefix * exponent)
        return
    elif unit == "kg":
        builder.add_base_unit_occurrence(SI_KILOGRAM, prefix * exponent)
        return
    elif unit == "s":
        builder.add_base_unit_occurrence(SI_SECOND, prefix * exponent)
        return
    elif unit == "A":
        builder.add_base_unit_occurrence(SI_AMPERE, prefix * exponent)
        return
    elif unit == "K":
        builder.add_base_unit_occurrence(SI_KELVIN, prefix * exponent)
        return
    elif unit == "mol":
        builder.add_base_unit_occurrence(SI_MOLE, prefix * exponent)
        return
    elif unit == "cd":
        builder.add_base_unit_occurrence(SI_CANDELA, prefix * exponent)
        return
    elif unit == "rad":
        # Follow boost units in defining radians as a base unit.
        builder.add_base_unit_occurrence(OTHER_RADIANS, prefix * exponent)
        return
    elif unit == "sr":
        # Follow boost units in defining steradian as a base unit.
        builder.add_base_unit_occurrence(OTHER_STERADIAN, prefix * exponent)
        return


    #
    # SI derived units:
    #
    if unit == "Pa":
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,   -1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -2 * prefix * exponent)
        return
    elif unit == "J":
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -2 * prefix * exponent)
        return
    elif unit == "W":
        # Watts
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -3 * prefix * exponent)
        return
    elif unit == "erg":
        # 1 erg = 1e-7 J
        builder.add_decadal_exponent(-7 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -2 * prefix * exponent)
        return
    elif unit == "g":
        builder.add_decadal_exponent(-3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        return
    elif unit == "h":
        builder.multiply_conversion_factor(pow(3600.0, prefix * exponent))
        builder.add_base_unit_occurrence(SI_SECOND, 1 * prefix * exponent)
        return
    elif unit == "Hz":
        # Hertz
        builder.add_base_unit_occurrence(SI_SECOND, -1 * prefix * exponent)
        return
    elif unit == "N":
        # Newton
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -2 * prefix * exponent)
        return
    elif unit == "W":
        # Watt
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -3 * prefix * exponent)
        return
    elif unit == "C":
        # Coulomb
        builder.add_base_unit_occurrence(SI_SECOND, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE, 1 * prefix * exponent)
        return
    elif unit == "V":
        # Volt
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,   -1 * prefix * exponent)
        return
    elif unit == "F":
        # Farad
        builder.add_base_unit_occurrence(SI_KILOGRAM, -1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    -2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,    4 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,    2 * prefix * exponent)
        return
    elif unit == "Ω":
        # Ohm, electrical resistance
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,   -2 * prefix * exponent)
        return
    elif unit == "S":
        # Siemens, electrical conductance
        builder.add_base_unit_occurrence(SI_KILOGRAM, -1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    -2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,    3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,    2 * prefix * exponent)
        return
    elif unit == "Wb":
        # Weber, magnetic flux
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,   -1 * prefix * exponent)
        return
    elif unit == "T":
        # Tesla, magnetic induction
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,   -1 * prefix * exponent)
        return
    elif unit == "H":
        # Henry, electrical inductance
        builder.add_base_unit_occurrence(SI_KILOGRAM,  1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,     2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,   -2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_AMPERE,   -2 * prefix * exponent)
        return
    elif unit == "lm":
        # Lumen, luminous flux
        builder.add_base_unit_occurrence(SI_CANDELA,      1 * prefix * exponent)
        builder.add_base_unit_occurrence(OTHER_STERADIAN, 1 * prefix * exponent)
        return
    elif unit == "lx":
        # Lux, illuminance
        builder.add_base_unit_occurrence(SI_CANDELA,      1 * prefix * exponent)
        builder.add_base_unit_occurrence(OTHER_STERADIAN, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,       -2 * prefix * exponent)
        return
    elif unit == "Bq":
        # Becquerel, radioactivity
        builder.add_base_unit_occurrence(SI_SECOND, -1 * prefix * exponent)
        return
    elif unit == "Gy" or unit == "Sv":
        # Gray, absorbed dose and
        # Sievert, equivalent dose
        builder.add_base_unit_occurrence(SI_METER,   2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND, -2 * prefix * exponent)
        return
    elif unit == "kat":
        # katal, catalyitic activity
        builder.add_base_unit_occurrence(SI_MOLE,    1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND, -1 * prefix * exponent)
        return
    elif unit == "bar":
        # bar, pressure
        builder.add_decadal_exponent(5 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,   -1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -2 * prefix * exponent)
        return
    elif unit == "l":
        # litre, volume
        builder.add_decadal_exponent(-3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER, 3 * prefix * exponent)
        return
    elif unit == "t":
        # ton, mass
        builder.add_decadal_exponent(3 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        return
    elif unit == "eV":
        # Electronvolt.
        # 1eV = 1.602176634e−19 J
        builder.multiply_conversion_factor(pow(1.602176634, prefix * exponent))
        builder.add_decadal_exponent(-19 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_KILOGRAM, 1 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_METER,    2 * prefix * exponent)
        builder.add_base_unit_occurrence(SI_SECOND,  -2 * prefix * exponent)
        return
    elif unit == "°C" or unit == "°F":
        raise ValueError("Cyantities does not support temperatures in Celsius "
            "or Fahrenheit since they are not proportional to the Kelvin scale. "
            "Please express your temperatures in Kelvin."
        )

    raise ValueError("Unknown unit '" + unit + "'")


cdef CppUnit parse_unit(str unit):
    """
    The central function that translates
    """
    # Early exit: Dimensionless, unit-unit:
    if unit == "1":
        return CppUnit()

    # Initialize the collected parsing results:
    cdef UnitBuilder builder

    # We allow the format like "kg*m/(s^2)"
    cdef list[str] nom_denom_split
    cdef str sub_unit
    cdef list nom_split
    cdef list denom_split
    if "/" in unit:
        # Fraction in the unit
        nom_denom_split = unit.split("/")
        if len(nom_denom_split) != 2:
            raise RuntimeError("Invalid format of unit '" + unit + "': "
                "only one division sign '/' allowed."
            )
        if nom_denom_split[1][0] != "(" or nom_denom_split[1][-1] != ")":
            raise RuntimeError("Invalid format of unit '" + unit + "': "
                "denominator needs to be enclosed in parantheses."
            )

        # Seems good to go.
        nom_split = nom_denom_split[0].split('*')
        denom_split = nom_denom_split[1][1:-1].split('*')

    else:
        # No fraction in the unit.
        if '*' in unit:
            nom_split = unit.split('*')
        else:
            nom_split = unit.split(' ')
        denom_split = []

    # Now add nominator and denominator:
    cdef size_t i
    cdef int exponent
    for sub_unit in nom_split:
        if sub_unit == "1":
            continue
        if "^" in sub_unit:
            sub_unit, exp = sub_unit.split("^")
            exponent = int(exp)
            if exponent == 0:
                continue
        else:
            exponent = 1
        _parse_unit_single(sub_unit, 1, exponent, builder)
    for sub_unit in denom_split:
        if "^" in sub_unit:
            sub_unit, exp = sub_unit.split("^")
            exponent = int(exp)
            if exponent <= 0:
                raise RuntimeError("Only positive exponents allowed if "
                                   "denominator is expressed through "
                                   "parantheses.")
        else:
            exponent = 1
        _parse_unit_single(sub_unit, -1, exponent, builder)

    return CppUnit(builder)


################################################################################
#                                                                              #
#                          Cython unit formatting                              #
#                                                                              #
################################################################################

cdef str _unit_id_to_string(base_unit_t uid):
    """
    Convert the 'base_unit_t' enum to the string representation
    of the unit.
    """
    if uid == SI_METER:
        return "m"
    elif uid == SI_KILOGRAM:
        return "kg"
    if uid == SI_SECOND:
        return "s"
    elif uid == SI_AMPERE:
        return "A"
    elif uid == SI_KELVIN:
        return "K"
    elif uid == SI_MOLE:
        return "mol"
    elif uid == SI_CANDELA:
        return "cd"
    elif uid == OTHER_RADIANS:
        return "rad"
    elif uid == OTHER_STERADIAN:
        return "sr"
    else:
        raise ValueError("Unknown base unit id")


cdef str format_unit(const CppUnit& unit, str rule):
    """
    Output a unit to a string using a specific formating
    rule.
    """
    # First get the scale:
    cdef double scale = unit.total_scale()
    cdef str s
    if scale != 1.0:
        s = str(scale) + " * "
    else:
        s = ""

    # Get the list of all units and exponents:
    cdef list olist = list()
    cdef base_unit_index_t i
    cdef int8_t occ
    for i in range(BASE_UNIT_COUNT):
        occ = unit.base_units()[i]
        olist.append((
            int(occ),
            _unit_id_to_string(_base_unit_from_index(i)),
            i
        ))

    # Now perform the different rules:
    if rule == 'coherent':
        s += " ".join(
            o[1] if o[0] == 1 else o[1] + "^" + str(o[0]) for o in olist
            if o[0] != 0
        )

    elif rule == 'casual':
        olist.sort(key = lambda o : (-o[0], o[2]))
        s += " ".join(
            o[1] if o[0] == 1 else o[1] + "^" + str(o[0]) for o in olist
            if o[0] != 0
        )

    else:
        raise RuntimeError("Rule '" + rule + "' not implemented.")

    return s


####################################################################################
#                                                                                  #
#                           The cdef Python class                                  #
#                                                                                  #
####################################################################################

cdef Unit generate_from_cpp(const CppUnit& unit):
    cdef Unit u = Unit.__new__(Unit)
    u._unit = unit
    return u

cdef Unit _multiply_units(Unit u0, Unit u1):
    return generate_from_cpp(u0._unit * u1._unit)


cdef Unit _divide_units(Unit u0, Unit u1):
    return generate_from_cpp(u0._unit / u1._unit)



cdef class Unit:
    """
    A physical unit.
    """
    def __init__(self, str unit):
        self._unit = parse_unit(unit)


    def __repr__(self) -> str:
        """
        String representation.
        """
        return "Unit(" + str(self) + ")"


    def format(self, rule='coherent') -> str:
        """
        Output this unit to a string using a specific formating
        rule.

        Parameters
        ----------
        rule : 'coherent' | 'casual'
        """
        if not isinstance(rule, str):
            raise TypeError("`rule` has to be a string!")
        return format_unit(self._unit, rule)


    def __str__(self) -> str:
        return self.format('coherent')


    def __mul__(self, other):
        cdef Unit result
        if isinstance(other, Unit):
            # Unit multiplied by Unit is a unit:
            return _multiply_units(self, other)

        return NotImplemented


    def __truediv__(self, other):
        cdef Unit result
        if isinstance(other, Unit):
            # Unit divided by Unit is a unit:
            return _divide_units(self, other)

        return NotImplemented


    def __pow__(self, exp):
        if not isinstance(exp, int):
            raise TypeError("Exponent needs to be integer in Unit power.")
        cdef int i = exp
        cdef Unit result = Unit.__new__(Unit)
        result._unit = self._unit.power(i)
        return result


    def __float__(self):
        # If dimensionless, we can convert to float:
        if self._unit.dimensionless():
            return self._unit.total_scale()

        # Attempting to lose dimension. Raise error!
        raise RuntimeError("Attempting to convert dimensional unit to float")


    def __eq__(self, other):
        if not isinstance(other, Unit):
            return False

        cdef Unit ou = other
        return self._unit == ou._unit


    def same_dimension(self, Unit other):
        return self._unit.same_dimension(other._unit)


    def dimensionless(self):
        """
        Queries whether this unit is dimensionless.
        """
        return self._unit.dimensionless()


# Quantities with units.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import numpy as np
from .unit cimport CppUnit, Unit
from libcpp cimport bool
from numpy cimport ndarray, float64_t


cdef extern from "cyantities/quantitywrap.hpp" namespace "cyantities" nogil:
    cppclass QuantityWrapper:
        QuantityWrapper()
        QuantityWrapper(double data, const CppUnit& unit)
        QuantityWrapper(double* data, size_t N, const CppUnit& unit)


cdef class Quantity:
    """
    A physical quantity: a single or array of real numbers with an associated
    physical unit.
    """
    # All underscored _variables are not part of the stable API.
    cdef bool _initialized
    cdef bool _is_scalar
    cdef double _val
    cdef size_t _val_array_N
    cdef double* _val_array_ptr
    # So as to hold a reference to the buffer, define the following:
    cdef object _val_object
    cdef CppUnit _unit
    cdef dict __dict__

    cdef _cyinit(self, bool is_scalar, double val, object val_object,
                 CppUnit unit)

    cdef QuantityWrapper wrapper(self) nogil

    @staticmethod
    cdef Quantity zeros_like(Quantity other, object unit)


# Run the parasolve code.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import numpy as np
import matplotlib.pyplot as plt

# The compiled model code:
from parasolve import ball_throw_with_friction

# Quantities to work with:
from cyantities import Quantity, Unit

#
# Model definition
# ================
#

t = Quantity(np.linspace(0.0, 6.0), 's')
x0 = Quantity(0.0, 'm')
y0 = Quantity(2.1, 'm')
v = Quantity(145.0, 'km h^-1')
vx0 = np.sin(np.pi/4) * v
vy0 = np.cos(np.pi/4) * v
r = Quantity(75.0, 'mm') / 2.0
cw = Quantity(0.45, '1')
rho = Quantity(0.875, 'g*cm^-3') # baseball
rho_air = Quantity(1.2, 'kg*m^-3')

#
# Model run
# =========
#
x,y = ball_throw_with_friction(t, x0, y0, vx0, vy0, r, cw, rho, rho_air)


#
# Plot the results
# ================
#
print(x / Unit('m'))
print(y / Unit('m'))
fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(np.array(x / Unit('m')), np.array(y / Unit('m')), marker='.')
ax.set_ylim(0, ax.get_ylim()[1])
ax.set_aspect('equal')
fig.savefig('result.pdf')


# SI Units in Python, accelerated by Cython.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libc.stdint cimport uint8_t, int8_t, int16_t
from libcpp cimport bool


#######################################################################
#
#                         C++ Unit backend.
#
#######################################################################

cdef extern from "cyantities/unit.hpp" namespace "cyantities":

    ctypedef uint8_t base_unit_index_t

    enum base_unit_t:
        SI_METER
        SI_KILOGRAM
        SI_SECOND
        SI_AMPERE
        SI_KELVIN
        SI_MOLE
        SI_CANDELA
        OTHER_RADIANS
        OTHER_STERADIAN

    const uint8_t BASE_UNIT_COUNT

    cdef cppclass base_unit_array_t:
        int8_t operator[](int) nogil

    struct UnitBuilder:
        int add_base_unit_occurrence(base_unit_t unit, int8_t exponent) nogil
        void add_decadal_exponent(int16_t exp) nogil
        void multiply_conversion_factor(double f) nogil


    cppclass CppUnit "cyantities::Unit":
        CppUnit() nogil
        CppUnit(int16_t dec_exp) nogil
        CppUnit(int16_t dec_exp, double conversion_factor) nogil
        CppUnit(const UnitBuilder&) nogil

        bool operator==(const CppUnit& other) nogil

        bool same_dimension(const CppUnit& other) nogil

        bool dimensionless() nogil

        CppUnit  operator*(CppUnit other) nogil

        CppUnit  operator/(CppUnit other) nogil

        CppUnit invert() nogil

        CppUnit power(int16_t exp) nogil

        int16_t decadal_exponent() nogil
        double conversion_factor() nogil
        double total_scale() nogil
        const base_unit_array_t& base_units() nogil



#######################################################################
#
#                              Cython.
#
#######################################################################

cdef CppUnit parse_unit(str unit)

cdef str format_unit(const CppUnit& unit, str rule)

cdef class Unit:
    cdef CppUnit _unit

cdef Unit generate_from_cpp(const CppUnit& unit)


# Build the parasolve code.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

import os
import glob
from shutil import copyfile
from pathlib import Path
from subprocess import run

#
# (1) Check where we are and where we need to go:
#
current = Path('.').resolve()
root = Path(__file__).resolve().parent

#
# (2) Meson build
#
os.chdir(root)
run(['meson','setup','builddir'])
os.chdir(root / "builddir")
run(['meson','compile'], check=True)

#
# (3) Copy the extension module:
#
# NOTE: This should not work on Windows.
# Probable fix: change the file extension to .dll
extname = glob.glob('parasolve*.so')[0]
os.chdir(root)
copyfile(root / "builddir" / extname, root / extname)


# Setup script.
#
# Author: Malte J. Ziebarth (mjz.science@fmvkb.de)
#
# Copyright (C) 2024 Malte J. Ziebarth
#
# Licensed under the EUPL, Version 1.2 or – as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

from pathlib import Path
from shutil import copyfile
from setuptools import setup, Extension
from mebuex import MesonExtension, build_ext

unit     = MesonExtension('cyantities.unit')
quantity = MesonExtension('cyantities.quantity')

#
# Post compile
# ------------
# Get the compiled static library for linking.
#
class InstallStaticLibrary(build_ext):
    """
    """
    def run(self):
        # First compile:
        super().run()

        # Copy the static library:
        builddir = Path('.') / "builddir"
        targetdir = quantity.mebuex_destpath
        copyfile(
            builddir / "libcyantities.a",
            targetdir / "_cpp" / "cyantities" / "libcyantities.a"
        )


setup(ext_modules=[unit, quantity],
      cmdclass={'build_ext' : InstallStaticLibrary}
)